import requests
from .higherLevelAPI import *
import itertools
from .dataCleaning import *
from .taxonomyAPI import *

def commonNameAPI(bestEffortTuple):
    name, speciesResult = bestEffortTuple
    if len(speciesResult) > 0: return (name,True,speciesResult)
    return commonNameReq(name)

def commonNameReq(name):
    try:
        res = requests.get("https://www.itis.gov/ITISWebService/jsonservice/searchByCommonName?srchKey="+name).json()
        commonNames = res['commonNames']
        if commonNames is None: return (name,False,[])
        if len(name.split(" ")) == 1: commonNames = list(filter(lambda x: checkIfCommonNameResultIsValid(name,x['commonName']),commonNames))
        commonNames = list(map(lambda x: (x['commonName'],x['tsn']) , commonNames))
        commonNames = list(map(tranlsateUidToSpeciesName,commonNames))
        return (name,True,list(set(map(cleanSingleSpeciesString,itertools.chain(*commonNames)))))
    except Exception as e:
        return (name,False,[])

def tranlsateUidToSpeciesName(dataTuple):
    commonName,uId = dataTuple
    res = requests.get("https://www.itis.gov/ITISWebService/jsonservice/getFullRecordFromTSN?tsn="+str(uId)).json()['scientificName']['combinedName']
    if len(res.strip().split(" ")) < 2: 
        return higherLevelAPI((commonName,('family',res)))[2]
    else:
        return [res]

def checkIfCommonNameResultIsValid(name,result):
    result = result.lower().split(" ")
    stringPotential1 = name 
    stringPotential2 = name+'s' #plurals
    return stringPotential1 in result or stringPotential2 in result



import re
from taxon_parser import TaxonParser, UnparsableNameException

def cleanHeadTailTupleData(headTailTuples):
    return list(filter(containsEmpty,map(cleanSingleTuple,headTailTuples)))

def containsEmpty(tup):
    return not(len(tup[0]) == 0 or len(tup[1]) == 0)

def cleanSingleTuple(tup):
    return (cleanSingleSpeciesString(tup[0]),cleanSingleSpeciesString(tup[1]))

def cleanSingleSpeciesString(species):
    species = species.lower()
    species = species.strip()
    species = species.replace("-"," ")
    species = re.sub(r'\{.*?\}', '', species)
    species = re.sub(r'\(.*?\)', '', species)
    species = re.sub(r'\[.*?\]', '', species)
    species = re.sub('[^a-zA-Z ]+',' ', species)
    species = re.sub(' sp ', ' ', species)  
    if species[-2:] == "sp":
        species = species[:-2]
    species = re.sub(r'\b\w{1}\b', '', species)
    species = re.sub(' +', ' ', species)  
    species = cleanNomenclatureFlags(species)
    species = species.lower()
    species = re.sub(' +', ' ', species)
    species = species.strip()  
    return species

def cleanNomenclatureFlags(species):
    if len(species.split(" ")) == 1: return species
    parser = TaxonParser(species)
    try: 
        parsed_name = parser.parse()
        if parsed_name.genus == "?":
            return parsed_name.specificEpithet + " " + parsed_name.infraspecificEpithet

        if not parsed_name.isTrinomial():
            species = parsed_name.genus + " " + parsed_name.getTerminalEpithet()
        else:
            species = parsed_name.genus + " " + parsed_name.specificEpithet
        return species
    except Exception as e:
        return species


import requests
from dataCleaning import *

def higherLevelAPI(nameAndTaxaTuple):
    name, taxaDetails = nameAndTaxaTuple
    if taxaDetails == '': return (name, False, [])
    if taxaDetails[0] == '': return (name,False,[])
    elif taxaDetails[0] == 'species': return (name,True,[cleanSingleSpeciesString(taxaDetails[1])])
    return convertHigherLevelTaxaToListOfSpecies(nameAndTaxaTuple)

def convertHigherLevelTaxaToListOfSpecies(nameAndTaxaTuple):
    name, taxaDetails = nameAndTaxaTuple
    taxaRank,taxaName = taxaDetails
    try:
        namesDicts = requests.get('https://api.gbif.org/v1/species/search?q='+taxaName+'&rank=species').json()['results']
        namesDicts = list(map(cleanSingleSpeciesString,map(lambda x: x.get('species',''), namesDicts)))
        namesDicts = list(filter(lambda x: x != '',namesDicts))
        if taxaRank == "genus": namesDicts = list(filter(lambda x: x.split(" ")[0] == taxaName,namesDicts))
        return (name,True,list(set(namesDicts)))
    except Exception as e:
        return (name,False,[])



import os
import pickle
import csv 
from .dataCleaning import *
from .taxonomyAPI import *
from .higherLevelAPI import *
from .commonNameAPI import *
import itertools
import pytest 

class EcoNameTranslator:
    def __init__(self):
        self.index = {}
        self.result = {}
        self.failed = {}
    
    def translate(self,uncheckedNames):
        for item in uncheckedNames:
            self.index[item] = [cleanSingleSpeciesString(item),'']
        self.tagInputWithTaxaLevel()
        self.mapKnownTaxaLevelToSpecies()
        self.tryCommonNameOnErrorCases()
        return self.index
    
    def tagInputWithTaxaLevel(self):
        names = list(map(lambda x: self.index[x][0], self.index.keys()))
        taxonomyAPIResuls, excepted = taxonomyAPI(names)
        self.writeApiResultsToIndex(taxonomyAPIResuls)
    
    def mapKnownTaxaLevelToSpecies(self):
        names = list(map(lambda x: self.index[x][0], self.index.keys()))
        taxaTuples = list(map(lambda x: self.index[x][1], self.index.keys()))
        originalNamesWthTaxa = list(zip(names,taxaTuples))
        speciesNamesOnly = list(map(higherLevelAPI,originalNamesWthTaxa))
        self.writeApiResultsToIndex(speciesNamesOnly)
    
    def tryCommonNameOnErrorCases(self):
        names = list(map(lambda x: self.index[x][0], self.index.keys()))
        currentSpeciesResults = list(map(lambda x: self.index[x][1], self.index.keys()))
        bestEffortOfSpecies = list(zip(names,currentSpeciesResults))
        speciesNamesOnly = list(map(commonNameAPI,bestEffortOfSpecies))
        self.writeApiResultsToIndex(speciesNamesOnly)
    
    def writeApiResultsToIndex(self,taxonomyAPIResuls):
        taxaAPIDict = {}
        for cleanedName,success,result in taxonomyAPIResuls:
            taxaAPIDict[cleanedName] = result
        
        for name in self.index:
            try: self.index[name][1] = taxaAPIDict[self.index[name][0]]
            except: pass
    
    def toPickleFile(self,path):
        with open(path,'wb') as fh:
            pickle.dump(index,fh) 
    
    def toCSV(self,path,name="translatedNames.csv"):
        headers = [["Input Name","Name After Cleaning","Result"]]
        csvData = map(lambda indivEcoName: \
                        list(map(lambda result: [indivEcoName,self.index[indivEcoName][0],result], \
                        self.index[indivEcoName][1])) \
                      ,self.index.keys())
                      
        csvData = list(itertools.chain(headers,*csvData))
        
        with open(path+"/"+name, "w", newline="") as f:
            writer = csv.writer(f)
            writer.writerows(csvData)




import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
     name='EcoNameTranslator',  
     version='0.3',
     author="Daniel Davies",
     author_email="dd16785@bristol.ac.uk",
     description="A package to translate ecological names in any format- from taxnomic rank (such as genus or family), or common English name (e.g. \"blackbird\") to specific, scientific species names.",
     long_description_content_type='text/markdown',
     long_description=long_description,
     url="https://github.com/Daniel-Davies/MedeinaTranslator",
     packages=setuptools.find_packages(),
     classifiers=[
         "Programming Language :: Python :: 3",
         "License :: OSI Approved :: MIT License",
         "Operating System :: OS Independent",
     ],
     install_requires=['taxon_parser'],
 )


import requests 

def taxonomyAPI(names):
    politeness = 50
    responses = []
    excepted = []
    for i in range(0,len(names),politeness):
        print(f"Processing Records {str(i)} to {str(min(i+politeness,len(names)))} [of {len(names)}]")
        apiInputWithCapitalisedGenus = list(map(lambda x: x.capitalize(),names[i:i+politeness]))
        succeeded,failed = callTaxaAPI(apiInputWithCapitalisedGenus)
        responses.extend(succeeded)
        excepted.extend(failed)

    speciesResponses = list(map(processSingleResponse,responses))
    return speciesResponses,excepted

def processSingleResponse(response):
    if response['is_known_name']: return (response['supplied_name_string'].lower(), True, parseSingleTaxonomyFromAPI(response))
    elif 'results' in response: return (response['supplied_name_string'].lower(),False, parseSingleTaxonomyFromAPI(response))
    return (response['supplied_name_string'], False, ('',''))

def parseSingleTaxonomyFromAPI(taxonomicAPIres):
    dataFromMultipleSources = taxonomicAPIres['results']
    dataFromMultipleSources = list(map(extractTaxaData,dataFromMultipleSources))
    for item in ['species','genus','subfamily','family']:
        for taxaMap in dataFromMultipleSources:
            if (item in taxaMap) and (taxaMap[item].strip() != ''): 
                return (item,taxaMap[item].strip())
    return ('','')
    
def extractTaxaData(singleTaxaSource):
    mappingDict = {}
    try: mappingDict = dict(zip(singleTaxaSource['classification_path_ranks'].lower().split("|"),\
                                singleTaxaSource['classification_path'].lower().split("|")))
    except: pass
    return mappingDict

def callTaxaAPI(rawNames):
    excepted = []
    preparedAPILoad = prepareTaxaAPIinput(rawNames) #try a bulk-call
    success,result = makeCall(preparedAPILoad) 
    if (not success) or (len(result) != len(rawNames)):
        namesSucceeded = list(filter(lambda resp: 'supplied_string_name' in resp, result))
        namesSucceeded = list(map(lambda resp: resp['supplied_string_name'], namesSucceeded))
        failedNames = list(set(rawNames) - set(namesSucceeded))
        succeededResults, failedNames = individuallyMakeRequests(failedNames)
        result.extend(succeededResults)
        excepted.extend(failedNames)
    
    return result, excepted

def prepareTaxaAPIinput(names):
    namesToSend = "|".join(names)
    return namesToSend
    
def individuallyMakeRequests(names):
    succeededResults = []
    failedNames = []
    for name in names:
        success,result = makeCall(name)
        if success: succeededResults.extend(result)
        else: failedNames.extend(name)
    
    return succeededResults, failedNames

def makeCall(preparedAPILoad):
    try:
        apiResult = requests.get(f'http://resolver.globalnames.org/name_resolvers.json?names={preparedAPILoad}').json()['data']
        return (True,apiResult)
    except Exception as e: 
        return (False,[])
